/***************************************************************************
                          kgraph.cpp  -  description
                             -------------------

    This file is a part of kpl - a program for graphical presentation of
    data sets and functions.

    begin                : Sun Apr 25 1999
    copyright            : (C) 2003 by Werner Stille
    email                : stille@uni-freiburg.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <math.h>
#include <qpaintdevicemetrics.h>
#include <qpixmap.h>
#include <kapp.h>
#include "kgraph.h"

//#define NO_CLIPPING

const char* KGraph::face[] = {"Helvetica", "Symbol", "itc zapf dingbats"};

KGraph::KGraph() : minx(0), maxx(0), miny(0), maxy(0), kxcur(0), kycur(0),
 ndigx(0), ndigy(0), minx0(0), miny0(0), pFactor(0), textdir(0.0),
 scalx(30.0), scaly(-30.0), xmin(0.0), xmax(15.0), ymin(0.0), ymax(10.0),
 xcur(0.0), ycur(0.0), dlen(15.0), dhite(10.0), xscal(30.0), yscal(-30.0),
 relang(0.0), relsin(0.0), relcos(1.0), dcur(0.0), factor(1.0),
 pen1(black, 0, SolidLine, RoundCap, RoundJoin),
 pen2(black, 0, SolidLine, RoundCap, RoundJoin),
 pen3(black, 0, SolidLine, RoundCap, RoundJoin)
{
  memset(dash, 0, sizeof(dash));
}

KGraph::~KGraph()
{
}

void KGraph::r2i(double x, double y, int* ix, int* iy)
{
  xcur = x;
  ycur = y;
  *ix = minx + qRound(xscal * (x - xmin));
  *iy = miny + qRound(yscal * (y - ymin));
}

bool KGraph::i2cm(const QPoint& p, double* xcm, double* ycm) const
{
  *xcm = (p.x() - minx0) / (scalx * factor);
  *ycm = (p.y() - miny0) / (scaly * factor);
  return ((p.x() < minx) || (p.x() > maxx) || (p.y() > miny) || (p.y() < maxy));
}

void KGraph::i2r(const QPoint& p, double _fxn, double _fyn,
                 double* xr, double* yr) const
{
  *xr = ((p.x() - minx) / xscal + xmin) / _fxn;
  if (logx)
    *xr = pow(10, *xr);
  *yr = ((p.y() - miny) / yscal + ymin) / _fyn;
  if (logy)
    *yr = pow(10, *yr);
}

QPoint KGraph::fromcm(double xcm, double ycm) const
{
  return QPoint(qRound(xcm * scalx * factor) + minx0,
                qRound(ycm * scaly * factor) + miny0);
}

void KGraph::cm2r(double xcm, double ycm, double _fxn, double _fyn,
                  double* xr, double* yr) const
{
  i2r(fromcm(xcm, ycm), _fxn, _fyn, xr, yr);
}

void KGraph::move(double x, double y)
{
  r2i(x, y, &kxcur, &kycur);
  moveTo(kxcur, kycur);
}

void KGraph::draw(double x, double y)
{
  if ((x != xcur) || (y != ycur)) {
    if (dash[1] == dash[0]) {
      r2i(x, y, &kxcur, &kycur);
      lineTo(kxcur, kycur);
    } else {
      dcur = dcur - dash[3] * (int) (dcur / dash[3]);
      double x1 = (x - xcur) * xscal / scalx;
      double y1 = (y - ycur) * yscal / scaly;
      double d0 = sqrt(x1 * x1 + y1 * y1);
      double d = d0;
      x1 = xcur;
      y1 = ycur;
      int i = 0;
      for (int j = 0; j < 3; j++)
        if (dcur >= dash[j])
          i = j + 1;
      while ((dcur + d) > dash[i]) {
        if (double f = (dash[i] - dcur) / d) {
          d = d * (1.0 - f);
          xcur = x - d * (x - x1) / d0;
          ycur = y - d * (y - y1) / d0;
          dcur = dash[i];
          if (i % 2)
            move(xcur,ycur);
          else {
            r2i(xcur, ycur, &kxcur, &kycur);
            lineTo(kxcur, kycur);
          }
        }
        if (i == 3)
          dcur = 0.0;
        i = ++i % 4;
      }
      xcur = x;
      ycur = y;
      if (i % 2)
        move(xcur, ycur);
      else {
        r2i(xcur, ycur, &kxcur, &kycur);
        lineTo(kxcur, kycur);
      }
      dcur += d;
    }
  }
}

void KGraph::line(double x1, double y1, double x2, double y2)
{
  move(x1, y1);
  draw(x2, y2);
}

void KGraph::plInit(const QPaintDevice* pd, int offLeft, int offTop)
{
  QPaintDeviceMetrics pdm(pd);
  begin(pd);
  eraseRect(window());
  if (pd->devType() == QInternal::Printer) {
    devType = Printer;
    pFactor = 4;
    QRect r = window();
    r.setSize(pFactor * r.size());
    setWindow(r);
  } else 
    pFactor = 1;
  fDpi = 75.0 / pdm.logicalDpiY();
  scalx = pFactor * pdm.logicalDpiX() / 2.54;
  scaly = -pFactor * pdm.logicalDpiY() / 2.54;
  minx0 = minx = -pFactor * offLeft;
  maxx = pFactor * pdm.width() + minx;
  miny0 = miny = pFactor * (pdm.height() + offTop);
  maxy = pFactor * offTop;
  logx = logy = false;
  xscal = scalx;
  yscal = scaly;
  dlen = (maxx - minx) / xscal;
  dhite = (maxy - miny) / yscal;
  factor = 1.0;
  xmin = ymin = 0.0;
  xcur = xmin;
  ycur = ymin;
  xmax = dlen;
  ymax = dhite;
  relang = 0.0;
  relsin = 0.0;
  relcos = 1.0;
  relSize = 1.0;
  textdir = 0.0;
  ndigx = ndigy = -1;
  linetype(1);
  pen1.setWidth((maxx - minx) / 240);
  pen2.setWidth((maxx - minx) / 480);
  pen3.setWidth((maxx - minx) / 339);
  setPen(pen3);
  QFont fnt = font();
  fnt.setFamily(face[0]);
  fnt.setPointSize(QMAX(qRound(0.0355 * fDpi * maxx), 1));
  setFont(fnt);
  move(xcur, ycur);
#ifndef NO_CLIPPING
  setClipRect(-offLeft, offTop, pdm.width(), pdm.height());
  setClipping(false);
#endif
}

void KGraph::format(double x, double y)
{
  xmin = ymin = 0.0;
  xmax = x;
  ymax = y;
  minx = minx0;
  miny = miny0;
  xscal = scalx * factor;
  yscal = scaly * factor;
  maxx = minx + qRound(xscal * xmax);
  maxy = miny + qRound(yscal * ymax);
  xcur = xmin;
  ycur = ymin;
#ifndef NO_CLIPPING
  setClipRect(minx / pFactor, maxy / pFactor, (maxx - minx) / pFactor,
              (miny - maxy) / pFactor);
  setClipping(false);
#endif
}

void KGraph::setColFrame(unsigned icol1)
{
  pen1.setColor(QColor(icol1));
}

void KGraph::setColGrid(unsigned icol2)
{
  pen2.setColor(QColor(icol2));
}

void KGraph::setColData(unsigned icol3)
{
  pen3.setColor(QColor(icol3));
  setPen(pen3);
}

void KGraph::setCol(unsigned icol1, unsigned icol2, unsigned icol3)
{
  setColFrame(icol1);
  setColGrid(icol2);
  setColData(icol3);
}

double KGraph::setRelSize(double r)
{
  double old = relSize;
  relSize = r;
  pen1.setWidth(qRound(0.00416667 * relSize * (maxx - minx)));
  pen2.setWidth(qRound(0.00208333 * relSize * (maxx - minx)));
  pen3.setWidth(qRound(0.00295 * relSize * (maxx - minx)));
  setPen(pen3);
  QFont fnt = font();
  fnt.setPointSize(QMAX(qRound(0.0355 * fDpi * relSize * (maxx - minx)), 1));
  setFont(fnt);
  return old;
}

void KGraph::Window(double xmi, double w, double ymi, double h)
{
  xmin = xmi;
  xmax = xmi + w;
  ymin = ymi;
  ymax = ymi + h;
  minx = minx0 + qRound(xscal * xmin);
  maxx = minx + qRound(xscal * w);
  miny = miny0 + qRound(yscal * ymin);
  maxy = miny + qRound(yscal * h);
  xcur = xmin;
  ycur = ymin;
  move(xcur, ycur);
  dcur = 0.0;
  int _dx = qRound(0.00416667 * relSize * (maxx - minx));
  pen1.setWidth(_dx);
  pen2.setWidth(qRound(0.00208333 * relSize * (maxx - minx)));
  pen3.setWidth(qRound(0.00295 * relSize * (maxx - minx)));
  setPen(pen3);
  QFont fnt = font();
  fnt.setPointSize(QMAX(qRound(0.0355 * fDpi * relSize * (maxx - minx)), 1));
  setFont(fnt);
#ifndef NO_CLIPPING
  setClipRect((minx + (_dx >> 1)) / pFactor , (maxy + (_dx >> 1)) / pFactor,
              (maxx - minx - _dx) / pFactor, (miny - maxy - _dx) / pFactor);
  setClipping(false);
#endif
}

void KGraph::scale(double xmi, double xma, double ymi, double yma,
                   bool lgx, bool lgy)
{
  logx = lgx;
  logy = lgy;
  xmin = logx ? log10(xmi) : xmi;
  xmax = logx ? log10(xma) : xma;
  ymin = logy ? log10(ymi) : ymi;
  ymax = logy ? log10(yma) : yma;
  xscal = (maxx - minx) / (xmax - xmin);
  yscal = (maxy - miny) / (ymax - ymin);
  xcur = xmin;
  ycur = ymin;
  move(xcur, ycur);
}

void KGraph::setDig(int nx, int ny)
{
  ndigx = nx;
  ndigy = ny;
}

void KGraph::frame()
{
  setPen(pen1);
  drawRect(minx, maxy, maxx - minx, miny - maxy);
  setPen(pen3);
}

void KGraph::raster(double xtic, double ytic, int intx, int inty, int mode,
                    double xOff, double yOff, bool xBottom, bool xTop,
                    bool yLeft, bool yRight)
{
  if (mode == Nothing)
    return;
  double dx = 0.0, dy, x, y;
  if (mode > FrameOnly) {
    if (logx) {
      if (xtic < 2.5)
        xtic = 2.0;
      else
        if (xtic < 5.5)
          xtic = 3.0;
        else
          if (xtic < 99.0)
            xtic = 10.0;
    }
    if (logy) {
      if (ytic < 2.5)
        ytic = 2.0;
      else
        if (ytic < 5.5)
          ytic = 3.0;
        else
          if (ytic < 99.0)
            ytic = 10.0;
    }
    setPen(pen2);
    dx = 0.99999 * (xmax - xmin);
    dy = 0.99999 * (ymax - ymin);
    if (mode <= AxesWithLabels) {
      dx *= 0.02;
      dy = -dx * xscal / yscal;
    }
    if (xOff && (!logx) && mode <= AxesWithLabels && intx > 1) {
      double d = xtic / intx;
      for (double xm = xmin + xOff - d; xm > xmin; xm -= d) {
        if (xBottom)
          line(xm, ymin, xm, ymin + 0.5 * dy);
        if (xTop)
          line(xm, ymax, xm, ymax - 0.5 * dy);
      }
    }
    for (x = xmin + xOff; x < xmax;)
      if (logx) {
        double xa = x;
        double mant = incLog(&x, xtic);
        if (x < xmax) {
          if (xBottom)
            line(x, ymin, x, ymin + dy);
          if (xTop)
            line(x, ymax, x, ymax - dy);
        }
        if (intx > 1) {
          double xd = pow(10.0, xa - mant);
          for (;;) {
            xa = log10(pow(10.0, xa) + xd);
            if ((xa >= x) || (xa >= xmax))
              break;
            if (xBottom)
              line(xa, ymin, xa,
                   ymin + (mode <= AxesWithLabels ? 0.5 * dy : dy));
            if (xTop)
              line(xa, ymax, xa,
                   ymax - (mode <= AxesWithLabels ? 0.5 * dy : dy));
          }
        }
      } else {
        if (mode <= AxesWithLabels && intx > 1) {
          double xm = x;
          for (int i = 1; i < intx; i++)
            if ((xm += xtic / intx) < xmax) {
              if (xBottom)
                line(xm, ymin, xm, ymin + 0.5 * dy);
              if (xTop)
                line(xm, ymax, xm, ymax - 0.5 * dy);
            }
        }
        if (x > xmin) {
          if (xBottom)
            line(x, ymin, x, ymin + dy);
          if (xTop)
            line(x, ymax, x, ymax - dy);
        }
        x += xtic;
      }
    if (yOff && (!logy) && mode <= AxesWithLabels && inty > 1) {
      double d = ytic / inty;
      for (double ym = ymin + yOff - d; ym > ymin; ym -= d) {
        if (yLeft)
          line(xmin, ym, xmin + 0.5 * dx, ym);
        if (yRight)
          line(xmax, ym, xmax - 0.5 * dx, ym);
      }
    }
    for (y = ymin + yOff; y < ymax;)
      if (logy) {
        double ya = y;
        double mant = incLog(&y, ytic);
        if (y < ymax) {
          if (yLeft)
            line(xmin, y, xmin + dx, y);
          if (yRight)
            line(xmax, y, xmax - dx, y);
        }
        if (inty > 1) {
          double yd = pow(10.0, ya - mant);
          for (;;) {
            ya = log10(pow(10.0, ya) + yd);
            if ((ya >= y) || (ya >= ymax))
              break;
            if (yLeft)
              line(xmin, ya,
                   xmin + (mode <= AxesWithLabels ? 0.5 * dx : dx), ya);
            if (yRight)
              line(xmax, ya,
                   xmax - (mode <= AxesWithLabels ? 0.5 * dx : dx), ya);
          }
        }
      } else {
        if (mode <= AxesWithLabels && inty > 1) {
          double ym = y;
          for (int i = 1; i < inty; i++)
            if ((ym += ytic / inty) < ymax) {
              if (yLeft)
                line(xmin, ym, xmin + 0.5 * dx, ym);
              if (yRight)
                line(xmax, ym, xmax - 0.5 * dx, ym);
            }
        }
        if (y > ymin) {
          if (yLeft)
            line(xmin, y, xmin + dx, y);
          if (yRight)
            line(xmax, y, xmax - dx, y);
        }
        y += ytic;
      }
  }
  if ((mode == AxesWithLabels) || (mode == GridWithLabels)) {
    QString label;
    double yB = 0.0, yT = 0.0;
    if (xBottom)
      yB = ymin + 0.03225 * relSize * (xmax - xmin) * xscal / yscal;
    if (xTop)
      yT = ymax - 0.03225 * relSize * (xmax - xmin) * xscal / yscal;
    int ix, iy;
    bool e = (logx && (xtic > 9.9) && ((xmax > 2.999999) || (xmin < -2.999999)));
    int nd = ndigx;
    double xe = xmax + 1.0e-3 * (xmax - xmin);
    for (x = xmin + xOff; x <= xe;) {
      if (x > xmax)
         x = xmax;
      if (e) {
        if (xBottom) {
          r2i(x, yB, &ix, &iy);
          text(label.sprintf("10#nu%d", qRound(x)), ix, iy, 2);
        }
        if (xTop) {
          r2i(x, yT, &ix, &iy);
          text(label.sprintf("10#nu%d", qRound(x)), ix, iy, 2);
        }
      } else {
        if (logx)
          if (x < 0)
            nd = 1 - int(x + 0.01);
          else
            nd = -1;
        int l = prenum(logx ? pow(10.0, x) : x, nd, label);
        setPen(pen1);
        if (xBottom) {
          r2i(x, yB, &ix, &iy);
          drawText(ix, iy, 0, 0, AlignCenter | DontClip, label, l);
        }
        if (xTop) {
          r2i(x, yT, &ix, &iy);
          drawText(ix, iy, 0, 0, AlignCenter | DontClip, label, l);
        }
      }
      if (logx)
        incLog(&x, xtic);
      else
        x += xtic;
    }
    double xL = 0.0, xR = 0.0;
    if (yLeft)
      xL = xmin - 0.025 * relSize * (xmax - xmin);
    if (yRight)
      xR = xmax + 0.025 * relSize * (xmax - xmin);
    dy = -0.0125 * relSize * (xmax - xmin) * xscal / yscal;
    e = (logy && (ytic > 9.9) && ((ymax > 2.999999) || (ymin < -2.999999)));
    nd = ndigy;
    double ye = ymax + 1.0e-3 * (ymax - ymin);
    for (y = ymin + yOff; y <= ye;) {
      if (y > ymax)
        y = ymax;
      if (e) {
        if (yLeft) {
          r2i(xL, y, &ix, &iy);
          text(label.sprintf("10#nu%d", qRound(y)), ix, iy, 3);
        }
        if (yRight) {
          r2i(xR, y, &ix, &iy);
          text(label.sprintf("10#nu%d", qRound(y)), ix, iy, 1);
        }
      } else {
        if (logy)
          if (y < 0)
            nd = 1 - int(y + 0.01);
          else
            nd = -1;
        int l = prenum(logy ? pow(10.0, y) : y, nd, label);
        setPen(pen1);
        if (yLeft) {
          r2i(xL, y, &ix, &iy);
          drawText(ix, iy - 1, 0, 0, AlignRight | AlignVCenter | DontClip,
                   label, l);
        }
        if (yRight) {
          r2i(xR, y, &ix, &iy);
          drawText(ix, iy - 1, 0, 0, AlignLeft | AlignVCenter | DontClip,
                   label, l);
        }
      }
      if (logy)
        incLog(&y, ytic);
      else
        y += ytic;
    }
  }
  setPen(pen1);
  frame();
  setPen(pen3);
}

void KGraph::setDir(double dir)
{
  textdir = dir;
  if (textdir) {
    relang = 0.01745329252 * textdir;
    relsin = sin(relang);
    relcos = cos(relang);
  } else {
    relang = relsin = 0.0;
    relcos = 1.0;
  }
}

void KGraph::text(const QString& str, int ix, int iy, int iorg, QRect* bRect)
{
  if (int l = str.length()) {
    QWMatrix wm = worldMatrix();
    translate(double(ix), double(iy));
    if (textdir)
      rotate(-textdir);
    setPen(pen1);
    int tf = 0;
    if (!str.contains('#')) {
      switch (iorg) {
        case 1:
          tf = AlignLeft | AlignVCenter | DontClip;
          break;
        case 2:
          tf = AlignCenter | DontClip;
          break;
        case 3:
          tf = AlignRight | AlignVCenter | DontClip;
      }
      drawText(0, 0, 0, 0, tf, str, -1, bRect);
    } else {
      const char fontcode[] = "NHBOSZI", levcode[] = "NUD";
      const int iface[7] = {0, 0, 0, 0, 1, 2, 0};
      const int weight[7] = {QFont::Normal, QFont::Normal, QFont::Bold,
                             QFont::Normal, QFont::Normal, QFont::Normal,
                             QFont::Bold};
      const bool italic[7] = {false, false, false, true, false, false, true};
      const double ys[2] = {0.6, -0.4};
      const char *c;
      int ix2 = 0;
      tf = AlignLeft | AlignVCenter | DontClip;
      double dr = 0.025 * relSize * (xmax - xmin) * xscal;
      int j0 = (iorg == 1) ? 1 : 0;
      int idx = 0;
      int idy;
      for (int j = j0; j < 2; j++) {
        if (j) {
          if (iorg == 2)
            idx >>= 1;
          if (iorg != 1)
            ix2 = -idx;
        }
        const char* s = str.latin1();
        int ilev = 0;
        while (*s) {
          if ((l = ((c = strchr(s, '#'))) ? (int) (c - s) : strlen(s))) {
            QRect r;
            if (j == 0) {
              r = boundingRect(0, 0, maxx, maxy,
                               AlignLeft | AlignVCenter, s, l);
              idx += r.width();
            } else {
              idy = (ilev) ? -qRound(dr * ys[ilev - 1]) : 0;
              drawText(ix2, idy, 0, 0, tf, s, l, &r);
              ix2 += r.width();
              if (bRect)
                *bRect = bRect->unite(r);
            }
          }
          if (c) {
            if (strlen(c) < 3)
              break;
            s = c + 3;
            char code;
            if ((code = *(++c)) > 96)
              code -= (char) 32;
            const char* f;
            if (!(f = strchr(fontcode, code)))
              break;
            int ifnt = (int) (f - fontcode);
            if ((code = *(++c)) > 96)
              code -= (char) 32;
            if (!(f = strchr(levcode, code)))
              break;
            setFont(QFont(face[iface[ifnt]],
                          QMAX(qRound(((ilev = (int) (f - levcode)) ?
                                       0.0251 : 0.0355) *
                                       relSize * fDpi * (maxx - minx)), 1),
#if KDE_VERSION_MAJOR > 2
                          weight[ifnt], italic[ifnt]));
#else
                          weight[ifnt], italic[ifnt],
                          iface[ifnt] ? QFont::AnyCharSet : QFont::Latin1));
#endif
          } else
            s += l;
        }
        setFont(QFont(face[iface[0]],
                QMAX(qRound(0.0355 * relSize * fDpi * (maxx - minx)), 1),
                weight[0], italic[0]));
      }
    }
    setPen(pen3);
    if (bRect)
      *bRect = xForm(*bRect);
    setWorldMatrix(wm);
  }
}

void KGraph::textcm(const QString& str, double x, double y, int al,
                    QRect* bRect)
{
  double xa, ya;
  double dh = 0.013 * relSize * (maxx - minx) / (scalx * factor);
  cm2r(x - dh * relsin, y + dh * relcos, 1.0, 1.0, &xa, &ya);
  int ix, iy;
  r2i(logx ? log10(xa) : xa, logy ? log10(ya) : ya, &ix, &iy);
  text(str, ix, iy, al, bRect);
}

void KGraph::arrow(double x, double y, double dir, double len,
                   QRect* bRect)
{
  QWMatrix wm = worldMatrix();
  QPoint p = fromcm(x, y);
  translate(double(p.x()), double(p.y()));
  if (dir)
    rotate(-dir);
  double dx = 0.03333 * relSize * (maxx - minx);
  int idx = -qRound(dx);
  int idy = QMAX(1, qRound(0.2 * dx));
  QPointArray a;
  a.setPoints(3, 0, 0, idx, idy, idx, -idy);
  setBrush(pen1.color());
  setPen(NoPen);
  drawPolygon(a);
  setBrush(NoBrush);
  setPen(pen1);
  int idx2 = -qRound(factor * scalx * len);
  if (idx2 <= idx)
    drawLine(idx, 0, idx2, 0);
  if (bRect)
    *bRect = xForm(QRect(idx2, -idy, -idx2, idy + idy));
  setPen(pen3);
  setWorldMatrix(wm);
}

void KGraph::arc(double x, double y, double w, double h, double a,
                 double alen, double dir, QRect* bRect)
{
  QWMatrix wm = worldMatrix();
  QPoint p = fromcm(x, y);
  translate(double(p.x()), double(p.y()));
  if (dir)
    rotate(-dir);
  int iw = qRound(factor * scalx * w);
  int ih = -qRound(factor * scaly * h);
  QRect r(-iw / 2, -ih / 2, iw, ih);
  drawArc(r, qRound(16 * a), qRound(16 * alen));
  if (bRect)
    *bRect = xForm(r);
  setWorldMatrix(wm);
}

void KGraph::rectangle(double x, double y, double w, double h,
                       bool filled, double dir, QRect* bRect)
{
  QWMatrix wm = worldMatrix();
  QPoint p = fromcm(x, y);
  translate(double(p.x()), double(p.y()));
  if (dir)
    rotate(-dir);
  int idy = qRound(factor * scaly * h);
  QRect r(0, idy, qRound(factor * scalx * w), -idy);
  if (filled) {
    setBrush(pen3.color());
    setPen(NoPen);
  }
  drawRect(r);
  if (filled) {
    setBrush(NoBrush);
    setPen(pen3);
  }
  if (bRect)
    *bRect = xForm(r);
  setWorldMatrix(wm);
}

void KGraph::ellipse(double x, double y, double w, double h,
                     bool filled, double dir, QRect* bRect)
{
  QWMatrix wm = worldMatrix();
  QPoint p = fromcm(x, y);
  translate(double(p.x()), double(p.y()));
  if (dir)
    rotate(-dir);
  int iw = qRound(factor * scalx * w);
  int ih = -qRound(factor * scaly * h);
  QRect r(-iw / 2, -ih / 2, iw, ih);
  if (filled) {
    setBrush(pen3.color());
    setPen(NoPen);
  }
  drawEllipse(r);
  if (filled) {
    setBrush(NoBrush);
    setPen(pen3);
  }
  if (bRect)
    *bRect = xForm(r);
  setWorldMatrix(wm);
}

void KGraph::letH(const QString& str, bool xTop)
{
  int ix, iy;
  r2i(0.5 * (xmin + xmax),
      ymax - (xTop ? 0.1875 : 0.04275) * relSize * (xmax - xmin) * xscal / yscal,
      &ix, &iy);
  text(str, ix, iy, 2);
}

void KGraph::letX(const QString& str, int iex, bool xTop)
{
  letxy(str, iex, 1, xTop);
}

void KGraph::letY(const QString& str, int iex, bool yRight)
{
  letxy(str, iex, 2, yRight);
}

void KGraph::setSymbol(int is)
{
  kSymb = is;
}

void KGraph::circle(double x0, double y0, double r)
{
  QPoint p = fromcm(x0, y0);
  int ir = qRound(factor * scalx * r);
  drawEllipse(p.x() - ir, p.y() - ir, ir + ir, ir + ir);
}

void KGraph::arcrel(double x, double y, double r, double angst, double angend)
{
  int x0, y0;
  r2i(x, y, &x0, &y0);
  int ir = qRound(xscal * fabs(r));
  double phi = 1.74532925e-2 * angst;
  int xstart = qRound(ir * cos(phi));
  int ystart = qRound(ir * sin(phi));
  if (r > 0.0) {
    x0 -= xstart;
    y0 += ystart;
  }
  drawArc(x0 - ir, y0 - ir, ir + ir, ir + ir,
          qRound(16.0 * angst), qRound(16.0 * angend));
}

double KGraph::incLog(double* x, double f)
{
  double mant = *x - int(*x + 1.0e-6 * fabs(*x));
  if (fabs(mant) < 1.0e-9)
    mant = 0.0;
  if (mant < 0.0)
    mant += 1.0;
  if (mant < 0.1)
    *x += log10(f);
  else
    if (mant < 0.4)
      *x += log10(2.5);
    else
      if (mant < 0.6)
        *x += log10(10.0 / 3.0);
      else
        *x += log10(2.0);
  return mant;
}

void KGraph::plArray(const double *x, const double *y, double fx, double fy,
                     int n, bool clip)
{
  if (kSymb == -17)
    return;
#ifndef NO_CLIPPING
  setClipping(clip);
#endif
  double xa, ya;
  if (kSymb > 0) {
    linetype(kSymb);
    xa = fx * x[0];
    ya = fy * y[0];
    bool b = false;
    if (((!logx) || (xa > 0)) && ((!logy) || (ya > 0))) {
      move(logx ? log10(xa) : xa, logy ? log10(ya) : ya);
      b = true;
    }
    for (int i = 1; i < n; i++) {
      xa = fx * x[i];
      ya = fy * y[i];
      if (((!logx) || (xa > 0)) && ((!logy) || (ya > 0)))
        if (b)
          draw(logx ? log10(xa) : xa, logy ? log10(ya) : ya);
        else {
          move(logx ? log10(xa) : xa, logy ? log10(ya) : ya);
          b = true;
        }
    }
    linetype(1);
  } else {
    int k = (-kSymb) % 17;
    QPointArray ap(10);
    if ((k > 10) || (k == 1)) {
      setBrush(pen3.color());
      setPen(NoPen);
    }
    const int ax[48] = {0, -2, 0, 2, 0, 2, -2, -2, 2, 2, 0, -2, 2, 0,
                        2, -2, 2, -2, 2, 2, 1, -1, -2, -2, -1, 1, 2, 2,
                        2, -2, 0, -2, 2, 0, 0, 0, -2, 2, 2, 0, -2, 0, 0,
                        0, -2, 2, 0, 0},
              ay[48] = {2, 0, -2, 0, 2, 2, 2, -2, -2, 2, 2, -1, -1, 2,
                        2, 2, -2, -2, 2, 1, 2, 2, 1, -1, -2, -2, -1, 1,
                        2, -2, 0, 2, -2, 2, -2, 0, 0, 0, 2, 0, 2, 0, -2,
                        2, 0, 0, 2, -2},
              as[10] = {0, 5, 10, 14, 19, 28, 33, 38, 43, 48};
    int j, l = 0, r;
    double f = 0.005 * relSize * (maxx - minx);
    for (int i = 0; i < n; i++) {
      xa = fx * x[i];
      ya = fy * y[i];
      if (((!logx) || (xa > 0)) && ((!logy) || (ya > 0))) {
        move(logx ? log10(xa) : xa, logy ? log10(ya) : ya);
        switch (k) {
          case 0:
          case 11:
            r = qRound(f + f);
            drawEllipse(kxcur - r, kycur - r, r + r, r + r);
            break;
          case 1:
            r = qRound(0.4 * f);
            drawEllipse(kxcur - r, kycur - r, r + r, r + r);
            break;
          default:
            r = (k < 12) ? k : k - 10;
            for (j = as[r - 2]; j < as[r - 1]; j++) {
              l = j - as[r - 2];
              if (k < 12)
                if (l)
                  lineTo(qRound(kxcur + f * ax[j]), qRound(kycur - f * ay[j]));
                else
                  moveTo(qRound(kxcur + f * ax[j]), qRound(kycur - f * ay[j]));
              else
                ap[l] = QPoint(qRound(kxcur + f * ax[j]), qRound(kycur - f * ay[j]));
            }
            if (k >= 12)
              drawPolygon(ap, false, 0, l);
        }
      }
    }
    if ((k > 10) || (k == 1)) {
      setBrush(NoBrush);
      setPen(pen3);
    }
  }
#ifndef NO_CLIPPING
  setClipping(false);
#endif
}

void KGraph::plError(const double* x, const double* y, const double* e,
                     double fx, double fy, int n)
{
#ifndef NO_CLIPPING
  setClipping(true);
#endif
  int idy = qRound(0.01 * relSize * (maxx - minx));
  int idx = idy >> 1;
  for (int i = 0; i < n; i++)
    if (((!logx) || (x[i] > 0)) && ((!logy) || (y[i] > 0))) {
      int ie;
      if (logy)
        ie = int(-0.4343 * e[i] * yscal / y[i]);
      else
        ie = int(-fy * e[i] * yscal);
      double xa = fx * x[i];
      if (logx)
        xa = log10(xa);
      double ya = fy * y[i];
      if (logy)
        ya = log10(ya);
      if ((xa >= xmin) && (xa <= xmax) &&
          (ya >= ymin) && (ya <= ymax) && (ie >= idy)) {
        move(xa, ya);
        moveTo(kxcur, kycur + idy);
        lineTo(kxcur, kycur + ie);
        moveTo(kxcur - idx, kycur + ie);
        lineTo(kxcur + idx, kycur + ie);
        moveTo(kxcur, kycur - idy);
        lineTo(kxcur, kycur - ie);
        moveTo(kxcur - idx, kycur - ie);
        lineTo(kxcur + idx, kycur - ie);
      }
    }
#ifndef NO_CLIPPING
  setClipping(false);
#endif
}

void KGraph::resetDash()
{
  dcur = 0.0;
}

void KGraph::linetype(int is)
{
  const int pat[9][2] = {{8, 8}, {7, 0}, {2, 0}, {1, 0}, {9, 0},
                         {8, 1}, {7, 2}, {4, 4}, {1, 1}};

  int j = QMAX(0, (is - 1) % 9);
  double f = relSize * (maxx - minx) / (scalx * 320.0);
  double dsp = (double) ((16 - pat[j][0] - pat[j][1]) >> 1);
  dash[0] = f * pat[j][0];
  dash[1] = dash[0] + f * dsp;
  dash[2] = dash[1] + f * pat[j][1];
  dash[3] = dash[2] + f * dsp;
}

void KGraph::letxy(const QString& str, int iex, int iax, bool opposite)
{
  if (str.length() || iex) {
    double x, y;
    QString label;
    double dx = 0.025 * relSize * (xmax - xmin);
    double dirsav = textdir;
    if (iax == 2) {
      int l = 3;
      if (logy)
        if ((ymax > 2.999999) || (ymin < -2.999999))
          l += (ymin < 0) ? 4 : 3;
        else
          l += 1 + int(QMAX(ymax, 1.99 - ymin));
      else
        l += QMAX(prenum(ymin, ndigy, label), prenum(ymax, ndigy, label));
      if (opposite)
        x = xmax + (0.75 * l + 0.71) * dx;
      else
        x = xmin - (0.75 * l + 0.71) * dx;
      y = 0.5 * (ymin + ymax);
      setDir(opposite ? -90.0 : 90.0);
    } else {
      x = 0.5 * (xmin + xmax);
      if (opposite)
        y = ymax - 3.79 * dx * xscal / yscal;
      else
        y = ymin + 3.79 * dx * xscal / yscal;
    }
    int ix, iy;
    r2i(x, y, &ix, &iy);
    if (iex)
      text("10#nu" + QString::number(iex) + "#nn " + str, ix, iy, 2);
    else
      text(str, ix, iy, 2);
    if (iax == 2)
      setDir(dirsav);
  }
}
